#include "command_buffers.h"
#include "device.h"

/* We need one command pool per frame per thread. */
#include <stdlib.h>
#include <volk/volk.h>
#include <runtime/runtime.h>
#include <runtime/atomics.h>

struct rt_vk_command_pool_set {
    VkCommandPool graphics;
    VkCommandPool compute;
    VkCommandPool transfer;
} ;

typedef struct {
    long indices[3];
} rt_thread_pool_indices;

static RT_THREAD_LOCAL rt_thread_pool_indices t_pool_indices;
static RT_THREAD_LOCAL bool t_are_indices_initialized[3] = {false, false, false};

rt_vk_command_pool_array rtReserveCommandPoolArray(long max_threads) {
    rt_vk_command_pool_array array;
    array.next_unused = 0;
    array.capacity = 0;
    array.pools = calloc(max_threads, sizeof(rt_vk_command_pool_set));
    if (!array.pools)
        return array;
    array.capacity = max_threads;
    return array;
}

void rtReleaseCommandPoolArray(rt_vk_command_pool_array *array) {
    free(array->pools);
    array->pools = NULL;
    array->capacity = 0;
    array->next_unused = 0;
}

static rt_vk_command_pool_set *GetThreadsPoolSet(rt_vk_device *dev, uint32_t frame_id) {
    uint32_t slot = frame_id % dev->max_frames_in_flight;
    if (!t_are_indices_initialized[slot]) {
        /* Alloc the set */
        rt_vk_command_pool_set *set = NULL;
        long index = rtAtomic32Inc(&dev->frames[slot].command_pools.next_unused) - 1; /* Inc returns the new value */
        if (index >= dev->frames[slot].command_pools.capacity) {
            rtReportError("VK", "Ran out of space for per-thread command buffers. Try to increase r_VkMaxThreads");
            return NULL;
        }
        set = &dev->frames[slot].command_pools.pools[index];

        /* Create the pools */
        {
            VkCommandPoolCreateInfo pool_info = {
                .sType = VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO,
                .flags = VK_COMMAND_POOL_CREATE_TRANSIENT_BIT,
                .queueFamilyIndex = dev->graphics_family,
            };
            if (vkCreateCommandPool(dev->device, &pool_info, dev->alloc_cb, &set->graphics) != VK_SUCCESS) {
                rtReportError("VK", "vkCreateCommandPool failed.");
                return NULL;
            }
#ifdef RT_DEBUG
            uint32_t thread_id = rtGetCurrentThreadId();
            char name[128];
            rtSPrint(name, 128, "Graphics Command Pool (FrameSlot %u, Thread %u)", slot, thread_id);
            VkDebugUtilsObjectNameInfoEXT name_info = {
                .sType        = VK_STRUCTURE_TYPE_DEBUG_UTILS_OBJECT_NAME_INFO_EXT,
                .objectHandle = (uint64_t)set->graphics,
                .objectType   = VK_OBJECT_TYPE_COMMAND_POOL,
                .pObjectName  = name,
            };
            vkSetDebugUtilsObjectNameEXT(dev->device, &name_info);
#endif
        }
        if (dev->compute_family != dev->graphics_family) {
            VkCommandPoolCreateInfo pool_info = {
                .sType = VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO,
                .flags = VK_COMMAND_POOL_CREATE_TRANSIENT_BIT,
                .queueFamilyIndex = dev->compute_family,
            };
            if (vkCreateCommandPool(dev->device, &pool_info, dev->alloc_cb, &set->compute) != VK_SUCCESS) {
                rtReportError("VK", "vkCreateCommandPool failed.");
                return NULL;
            }
#ifdef RT_DEBUG
            uint32_t thread_id = rtGetCurrentThreadId();
            char name[128];
            rtSPrint(name, 128, "Compute Command Pool (FrameSlot %u, Thread %u)", slot, thread_id);
            VkDebugUtilsObjectNameInfoEXT name_info = {
                .sType        = VK_STRUCTURE_TYPE_DEBUG_UTILS_OBJECT_NAME_INFO_EXT,
                .objectHandle = (uint64_t)set->compute,
                .objectType   = VK_OBJECT_TYPE_COMMAND_POOL,
                .pObjectName  = name,
            };
            vkSetDebugUtilsObjectNameEXT(dev->device, &name_info);
#endif
        } else {
            set->compute = set->graphics;
        }
        if (dev->transfer_family != dev->graphics_family && dev->transfer_family != dev->compute_family) {
            VkCommandPoolCreateInfo pool_info = {
                .sType = VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO,
                .flags = VK_COMMAND_POOL_CREATE_TRANSIENT_BIT,
                .queueFamilyIndex = dev->transfer_family,
            };
            if (vkCreateCommandPool(dev->device, &pool_info, dev->alloc_cb, &set->transfer) != VK_SUCCESS) {
                rtReportError("VK", "vkCreateCommandPool failed.");
                return NULL;
            }
#ifdef RT_DEBUG
            uint32_t thread_id = rtGetCurrentThreadId();
            char name[128];
            rtSPrint(name, 128, "Transfer Command Pool (FrameSlot %u, Thread %u)", slot, thread_id);
            VkDebugUtilsObjectNameInfoEXT name_info = {
                .sType        = VK_STRUCTURE_TYPE_DEBUG_UTILS_OBJECT_NAME_INFO_EXT,
                .objectHandle = (uint64_t)set->transfer,
                .objectType   = VK_OBJECT_TYPE_COMMAND_POOL,
                .pObjectName  = name,
            };
            vkSetDebugUtilsObjectNameEXT(dev->device, &name_info);
#endif
        }
        else if (dev->transfer_family == dev->graphics_family) {
            set->transfer = set->graphics;
        } else {
            RT_ASSERT(dev->transfer_family == dev->compute_family, "");
            set->transfer = set->compute;
        }

        t_pool_indices.indices[slot] = index;
        t_are_indices_initialized[slot] = true;
    }

    long index = t_pool_indices.indices[slot];
#ifdef RT_DEBUG
    if (index >= dev->frames[slot].command_pools.next_unused)
        return NULL;
#endif
    return &dev->frames[slot].command_pools.pools[index];
}

VkCommandPool rtGetGraphicsCommandPool(rt_vk_device *dev) {
    rt_vk_command_pool_set *set = GetThreadsPoolSet(dev, dev->current_frame_id);
    return RT_VERIFY(set)->graphics;
}

VkCommandPool rtGetComputeCommandPool(rt_vk_device *dev) {
    rt_vk_command_pool_set *set = GetThreadsPoolSet(dev, dev->current_frame_id);
    return RT_VERIFY(set)->compute;
}

VkCommandPool rtGetTransferCommandPool(rt_vk_device *dev) {
    rt_vk_command_pool_set *set = GetThreadsPoolSet(dev, dev->current_frame_id);
    return RT_VERIFY(set)->transfer;
}

VkCommandBuffer rtAllocateGraphicsCommandBuffer(rt_vk_device *dev) {
    VkCommandPool pool = rtGetGraphicsCommandPool(dev);
    VkCommandBufferAllocateInfo alloc_info = {
        .sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO,
        .commandPool =  pool,
        .level = VK_COMMAND_BUFFER_LEVEL_PRIMARY,
        .commandBufferCount = 1,
    };
    VkCommandBuffer cmdbuf = VK_NULL_HANDLE;
    vkAllocateCommandBuffers(dev->device, &alloc_info, &cmdbuf);
    return cmdbuf;
}

VkCommandBuffer rtAllocateComputeCommandBuffer(rt_vk_device *dev) {
    VkCommandPool pool = rtGetComputeCommandPool(dev);
    VkCommandBufferAllocateInfo alloc_info = {
        .sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO,
        .commandPool =  pool,
        .level = VK_COMMAND_BUFFER_LEVEL_PRIMARY,
        .commandBufferCount = 1,
    };
    VkCommandBuffer cmdbuf = VK_NULL_HANDLE;
    vkAllocateCommandBuffers(dev->device, &alloc_info, &cmdbuf);
    return cmdbuf;
}

VkCommandBuffer rtAllocateTransferCommandBuffer(rt_vk_device *dev) {
    VkCommandPool pool = rtGetTransferCommandPool(dev);
    VkCommandBufferAllocateInfo alloc_info = {
        .sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO,
        .commandPool =  pool,
        .level = VK_COMMAND_BUFFER_LEVEL_PRIMARY,
        .commandBufferCount = 1,
    };
    VkCommandBuffer cmdbuf = VK_NULL_HANDLE;
    vkAllocateCommandBuffers(dev->device, &alloc_info, &cmdbuf);
    return cmdbuf;
}